// =============== Klassen-Definition: BASIS  ===========================================
// Tensor-Klasse BASIS fuer Basisvektoren, Metriken, Volumenelmente und epsilon-Tensoren
// Basisvektoren g(i,alpha) und Matrizen mat(i,j,k,l) werden spaltenweise gespeichert !!!
// Indizierung von alpha,i,j,k,l beginnt bei 0 !!!
// vgl. U. Breymann, C++, Carl Hanser Verlag 1993, S.186 ff
// ======================================================================================
// Copyright: Univ.-Prof. Dr.-Ing. Udo F. Meissner, Erkrath/Darmstadt
// Lehr- und Studienprogramm - kommerzielle Nutzung nicht gestattet. 
// Der Autor uebernimmt keine Haftung fuer Unvollkommenheit und Fehler 
// Version:  1996/2022
// Compiler: C++
// ================ Include-Files =======================================================
#include "pch.h"
#include "framework.h"
#include "BASISclass.h"
//
// #include "Basis.h"
// #include "MATRIX.h"
// #include "VEKTOR.h"
// ============================================================================
// public -> Oeffentliche Mitgliedsfunktionen
// ============================================================================
//
// Konstruktor_0 -> BASIS g(mat)
// =============================
// mat: Matrix mit kovarianten Basisvektoren
// -----------------------------------------
BASIS::BASIS(MATRIX &mat)
{
//***cout<<"BASIS::Konstruktor_0: neue Basis aus MATRIX mat erzeugen"<<'\n'; 
//
//
// Dimensionen ni,nbeta von MATRIX mat(ni,nbeta) ermitteln
// -------------------------------------------------------
int n;
//
mat.dimension_ermitteln(ni,nbeta,n,n);
//
if (ni < nbeta) cout << " Fehler in BASIS-Konstruktor_0: g[ni,nbeta] Dimension ni =" 
                     << ni << " < nbeta= " << nbeta << " !!!" << endl;
//
this->erzeugen(mat);
//
//***cout<<"BASIS::Konstruktor_0: Matrix mit Basisvektoren "<<mat;
//
//***cout<<"BASIS::Konstruktor_0: Ende"<<'\n';
};//---------------------------------------------------------------------------
//
// Konstruktor_1 -> BASIS g(vek1)
// ==============================
// vek1: Vektor mit kovariantem Basisvektor 1
// ------------------------------------------
BASIS::BASIS(VEKTOR &vek1)
{
//***cout<<"BASIS::Konstruktor_1: neue Basis aus VEKTOR vek1 erzeugen"<<'\n'; 
//
// Dimensionen ni,nbeta von MATRIX mat(ni,nbeta) ermitteln
// -------------------------------------------------------
nbeta=1;
//
int n;
vek1.dimension_ermitteln(n,ni,n,n);
//
MATRIX &mat=*new MATRIX(ni,nbeta);
//
if (ni < nbeta) cout << " Fehler in BASIS-Konstruktor_1: g[ni,nbeta] Dimension ni =" 
                     << ni << " < nbeta= " << nbeta << " !!!" << endl;
//
// Vektor in Matrix einspeichern
mat=vek1;
//
//***cout<<"BASIS::Konstruktor_1: Matrix mit Basisvektoren "<<mat;
//
this->erzeugen(mat);
//
// Matrix loeschen
delete &mat;
//
//***cout<<"BASIS::Konstruktor_1: Ende"<<'\n';
};//---------------------------------------------------------------------------
//
// Konstruktor_2 -> BASIS g(vek1,vek2)
// ===================================
// vek1: Vektor mit kovariantem Basisvektor 1
// vek2: Vektor mit kovariantem Basisvektor 2
// ------------------------------------------
BASIS::BASIS(VEKTOR &vek1,VEKTOR &vek2)
{
//***cout<<"BASIS::Konstruktor_2: neue Basis aus VEKTOR vek1,vek2 erzeugen"<<'\n'; 
//
//
// Dimensionen ni,nbeta von MATRIX mat(ni,nbeta) ermitteln
// -------------------------------------------------------
nbeta=2;
//
int n,mi;
vek1.dimension_ermitteln(n,ni,n,n);
vek2.dimension_ermitteln(n,mi,n,n);
if(mi > ni) ni=mi;
//
MATRIX &mat=*new MATRIX(ni,nbeta);
//
if (ni < nbeta) cout << " Fehler in BASIS-Konstruktor_2: g[ni,nbeta] Dimension ni =" 
                     << ni << " < nbeta= " << nbeta << " !!!" << endl;
//
// Vektor in Matrix einspeichern
mat=vek1;
mat.teilmatrix_hinein(0,1)=vek2;
//
//***cout<<"BASIS::Konstruktor_2: Matrix mit Basisvektoren "<<mat;
//
this->erzeugen(mat);
//
// Matrix loeschen
delete &mat;
//
//***cout<<"BASIS::Konstruktor_2: Ende"<<'\n';
};//---------------------------------------------------------------------------
//
// Konstruktor_3 -> BASIS g(vek1,vek2,vek3)
// ========================================
// vek1: Vektor mit kovariantem Basisvektor 1
// vek2: Vektor mit kovariantem Basisvektor 2
// vek3: Vektor mit kovariantem Basisvektor 3
// ------------------------------------------
BASIS::BASIS(VEKTOR &vek1,VEKTOR &vek2,VEKTOR &vek3)
{
//***cout<<"BASIS::Konstruktor_3: neue Basis aus "<<
//***                                   "VEKTOR vek1,vek2,vek3 erzeugen"<<'\n'; 
//
//
// Dimensionen ni,nbeta von MATRIX mat(ni,nbeta) ermitteln
// -------------------------------------------------------
nbeta=3;
//
int n,mi;
vek1.dimension_ermitteln(n,ni,n,n);
vek2.dimension_ermitteln(n,mi,n,n);
if(mi > ni) ni=mi;
vek3.dimension_ermitteln(n,mi,n,n);
if(mi > ni) ni=mi;
//
MATRIX &mat=*new MATRIX(ni,nbeta);
//
if (ni < nbeta) cout << " Fehler in BASIS-Konstruktor_3: g[ni,nbeta] Dimension ni =" 
                     << ni << " < nbeta= " << nbeta << " !!!" << endl;
//
// Vektor in Matrix einspeichern
mat=vek1;
mat.teilmatrix_hinein(0,1)=vek2;
mat.teilmatrix_hinein(0,2)=vek3;
//
//***cout<<"BASIS::Konstruktor_3: Matrix mit Basisvektoren "<<mat;
//
this->erzeugen(mat);
//
// Matrix loeschen
delete &mat;
//
//***cout<<"BASIS::Konstruktor_3: Ende"<<'\n';
};//---------------------------------------------------------------------------
//
// Destruktor -> delete &g
// =======================
BASIS::~BASIS(void)
{
//***cout<<"BASIS::Destruktor: Basis loeschen"<<'\n';
//
// Loeschen erlauben
g_kovariant->entsperren();
delete g_kovariant;
g_kontravariant->entsperren();
delete g_kontravariant;
//
//***cout<<"BASIS::Destruktor: Basis geloescht"<<'\n';
};//---------------------------------------------------------------------------
//
// Kovariante Basisvektoren -> g.kovariante_basis()
// ================================================
//
MATRIX &BASIS::kovariante_basis()
{
//***cout<<"Kovariante Basisvektoren "<<*g_kovariant;
//
// gegen Loeschen sperren
g_kovariant->sperren();
return *g_kovariant;
};//---------------------------------------------------------------------------
//
// Kontravariante Basisvektoren -> g.kontravariante_basis()
// ========================================================
//
MATRIX &BASIS::kontravariante_basis()
{
//***cout<<"Kontravariante Basisvektoren "<<*g_kontravariant;
//
// gegen Loeschen sperren
g_kontravariant->sperren();
return *g_kontravariant;
};//---------------------------------------------------------------------------
//
// Laengen- bzw. Flaechen- bzw. Volumen-Element -> g.wurzel_g()
// ============================================================
// wurzel_g = sqrt(Determinate von g(alpha,beta))
// ----------------------------------------------
double &BASIS::wurzel_g()
{
//***cout<<"Laengen-/Flaechen-/Volumen-Element= "<<g_wurzel<<'\n';
//
return g_wurzel;
};//---------------------------------------------------------------------------
//
// Kovariante Metrik -> g.kovariante_metrik(mat)
// =============================================
// g(alpha,beta) = g(i,alpha) * g(i,beta)
// Ergebnis: Determinante (g)
// --------------------------------------
// 
double BASIS::kovariante_metrik(MATRIX &mat)
{
mat=g_kovariant->dimension(1,ni,nbeta)*g_kovariant->dimension(1,ni,nbeta);
//
//***cout<<" BASIS: Kovariante Metrik "<< mat;
//
// Determinante g der Metrik mat emitteln
// ======================================
double g = mat.determinante(nbeta);
//
//***cout<<" BASIS: Determinante  g = "<< g;
//
// im Datenbereich speichern
g_wurzel = sqrt(g);
//
// Warnmeldung (falls Basisvektoren (Zeilen & Spalten der Metrik) 0 sind,
//              wird das bei der kontravarianten Metrik abgefangen,
//              weil diese im Gleichungssystem als Lagerbedingung gelten !)
if (g == 0.0)
{cout << " Warnung der kovarianten Metrik: Determinante g = " << g << endl;
 cout << " kovariante Basisvektoren sind linear abhaengig oder 0 !!! " << *g_kovariant << endl;
 cout << " 0-Spalten sind bei Bildung der kontravarianten Metrik jedoch als Lagerung beruecksichtigt ! " << endl;
} 
return g;
};//---------------------------------------------------------------------------
//
// Kontravariante Metrik -> g.kontravariante_metrik(mat)
// =====================================================
// g(alpha,beta) = g(i,alpha) * g(i,beta)
// --------------------------------------
void BASIS::kontravariante_metrik(MATRIX &mat)
{
mat=
 g_kontravariant->dimension(1,ni,nbeta)*g_kontravariant->dimension(1,ni,nbeta);
//
//***cout<<"Kontravariante Metrik "<<mat;
};//---------------------------------------------------------------------------
//
// Kovarianter epsilon-Tensor -> g.kovarianter_epsilon_tensor(mat)
// ===============================================================
void BASIS::kovarianter_epsilon_tensor(MATRIX &mat)
{
int dim2=1,dim3=1;
if(nbeta >= 2) dim2=nbeta;
if(nbeta == 3) dim3=nbeta;
//
mat=*new e_MATRIX(nbeta,dim2,dim3,g_wurzel);
//
//***cout<<"Kovarianter epsilon-Tensor "<<mat;
};//---------------------------------------------------------------------------
//
//
// Kontravarianter epsilon-Tensor -> g.kontravarianter_epsilon_tensor(mat)
// =======================================================================
//
void BASIS::kontravarianter_epsilon_tensor(MATRIX &mat)
{
int dim2=1,dim3=1;
if(nbeta >= 2) dim2=nbeta;
if(nbeta == 3) dim3=nbeta;
//
mat=*new e_MATRIX(nbeta,dim2,dim3,1.0/g_wurzel);
//
//***cout<<"Kontravarianter epsilon-Tensor "<<mat;
};//---------------------------------------------------------------------------
// 
// ===========================================================================
//                 Private Mitgliedsfunktionen
//                 ===========================
//
// Basis erzeugen -> erzeugen(mat)
// ===============================
// mat: Matrix mit kovarianten Basisvektoren
// -----------------------------------------
void BASIS::erzeugen(MATRIX &mat)
{
 bool FEHLER;
 FEHLER = false;
//*** FEHLER=true;
//*** cout<<"BASIS::erzeugen: Basis erzeugen "<<'\n';
//
// kovariante Basisvektoren
// ------------------------
g_kovariant = new MATRIX(ni,nbeta);
*g_kovariant = mat;
//
// *** cout<<"BASIS::erzeugen: Kovariante Basisvektoren "<<*g_kovariant;
//
// kovariante Metrik
// -----------------
MATRIX &metrik_kovariant=*new MATRIX(nbeta,nbeta);
// metrik_kovariant(alpha,beta) = g_kovariant(i,alpha) * g_kovariant(i,beta)
//
this->kovariante_metrik(metrik_kovariant);
//***cout<<"BASIS::erzeugen: Kovariante Metrik "<<metrik_kovariant;
// 
// kontravariante Metrik
// ---------------------
MATRIX &metrik_kontravariant=*new MATRIX(nbeta,nbeta);
// Inverse der kovarianten Metrik bilden
g_wurzel=metrik_kovariant.invertieren(metrik_kontravariant,1);

//***cout<<"BASIS::erzeugen: Kontravariante Metrik "<<metrik_kontravariant;
//
// Laengen- bzw. Flaechen- bzw. Volumenelement
// -------------------------------------------
g_wurzel=sqrt(g_wurzel);
//***cout<<"BASIS::erzeugen: Laengen-/Flaechen-/Volumen-Element= "<<g_wurzel<<'\n';
//
// kontravariante Basisvektoren
// ----------------------------
g_kontravariant=new MATRIX(ni,nbeta);
//
//g_kontravariant(i,beta)=g_kovariant(i,alpha)*metrik_kontravariant(alpha,beta)
//
 *g_kontravariant=(*g_kovariant)*metrik_kontravariant.dimension(1,nbeta,nbeta);
//
//***cout<<"BASIS::erzeugen: Kontravariante Basisvektoren "<<*g_kontravariant;
//
// ----------------------------------------------------------------------------
// Probe: delta(alpha,beta) = g_kovariant(i,alpha) * g_kontravariant(i,beta)
// -------------------------------------------------------------------------
if(FEHLER == true)
{
MATRIX &delta=*new MATRIX(nbeta,nbeta);
delta=g_kovariant->dimension(1,ni,nbeta)*g_kontravariant->dimension(1,ni,nbeta);
cout<<"BASIS::erzeugen: Probe:"<<'\n'
    <<"       Einheitsmatrix = g_kovariant * g_kontravariant "<<delta;
delete &delta;
};
// ----------------------------------------------------------------------------
//
delete &metrik_kovariant;
delete &metrik_kontravariant;
};
//---------------------------------------------------------------------------
